/*
 * Echo Client Programm.. Hacked von Ewan Birney <birney@sanger.ac.uk>
 * aus der echo test suite, aktualisiert für ORBit2 von Frank Rehberger
 * <F.Rehberger@xtradyne.de>
 *
 * Der Client ließt die Object-Referenz (IOR) aus der lokalen Datei 'echo.ref'
 * und leitet Konsoleneingaben zum echo-server weiter. Ein Punkt . als einziges
 * Zeichen in der Eingabe beendet den client.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <orbit/orbit.h>

/*
 * Die folgende Headerdatei wurde durch die idl erstellt.
 */

#include "echo.h"

/**
 * auf Ausnahme (exception) testen
 */
static
gboolean
raised_exception(CORBA_Environment *ev)
{
   return ((ev)->_major != CORBA_NO_EXCEPTION);
}

/**
 * falls eine Ausnahme aufgetreten ist beendet diese Funktion den Prozess
 */
static
void
abort_if_exception(CORBA_Environment *ev, const char* mesg)
{
   if (raised_exception (ev)) {
      g_print ("%s %s", mesg, CORBA_exception_id (ev));
      CORBA_exception_free (ev);
      abort();
   }
}

static CORBA_ORB global_orb = CORBA_OBJECT_NIL;    /* global orb */

/*
 * Folgendes wird aufgerufen falls Prozess-Signale auftreten. Es ruft
 * CORBA_ORB_shutdown() auf, welches die Prozesshauptschleife beendet.
 */
static
void
client_shutdown (int sig)
{
   CORBA_Environment local_ev[1];
   CORBA_exception_init(local_ev);

   if(global_orb != CORBA_OBJECT_NIL)
   {
      CORBA_ORB_shutdown (global_orb, FALSE, local_ev);
      abort_if_exception (local_ev, "caught exception");
   }
}

/*
 * Initialisiert den ORB @orb durch die @argv Parameter. Für jeden ORBit
 * Parameter der aus @argv benutzt wird, wird der @argc_ptr um eins
 * verringert. Der Signal-Handler wird eingerichtet, so daß er
 * echo_client_shutdown aufruft, wenn die Signale SIGINT oder SIGTERM auftreten.
 * Wenn ein Fehler auftritt zeigt @ev nach der Rückkehr auf ein Exception
 * Objekt.
 */
static 
void
client_init (int                 *argc_ptr,
             char                *argv[],
             CORBA_ORB           *orb,
             CORBA_Environment   *ev)
{
   /* initialisiere Signalhandling */
   
   signal (SIGINT,   client_shutdown);
   signal (SIGTERM,  client_shutdown);

   /* erzeuge Object Request Broker (ORB) */

   (*orb) = CORBA_ORB_init (argc_ptr, argv, "orbit-local-mt-orb", ev);
   if (raised_exception (ev))
      return;
}

/*
 * Releases @servant Objekt und zerstört schlußendlich @orb. Wenn ein Fehler
 * auftritt, zeigt @ev nach der Rückkehr auf ein Exception Objekt.
 */
static
void
client_cleanup (CORBA_ORB           orb,
                CORBA_Object        service,
                CORBA_Environment   *ev)
{
   /* verwaltetes Objekt freigeben */
   CORBA_Object_release (service, ev);
   if (raised_exception (ev))
      return;

   /* den ORB herunterfahren */
   if (orb != CORBA_OBJECT_NIL)
   {
      /* versuche den ORB zu zerstören.. */
      CORBA_ORB_destroy (orb, ev);
      if (raised_exception (ev))
         return;
   }
}

/**
 *
 */
static
CORBA_Object
client_import_service_from_stream (CORBA_ORB          orb,
                                   FILE               *stream,
                                   CORBA_Environment  *ev)
{
   CORBA_Object obj = CORBA_OBJECT_NIL;
   CORBA_char objref[4096];

   g_print ("DEBUG: before fscanf\n");
   fgets (objref, 4096, stream);    /* FIXME, handle input errror */
   g_print ("DEBUG: before fscanf\n");

   g_print ("IOR is: %s\n", objref);

   obj = (CORBA_Object) CORBA_ORB_string_to_object (global_orb,
                                                    objref,
                                                    ev);

   CORBA_free (objref);

   return obj;
}

/**
 *
 */
static
CORBA_Object
client_import_service_from_file (CORBA_ORB            orb,
                                 char                 *filename,
                                 CORBA_Environment    *ev)
{
   CORBA_Object   obj   = NULL;
   FILE           *file = NULL;

   file = fopen (filename, "r");
   if (file == NULL)
      g_error ("could not open %s\n", filename);

   obj = client_import_service_from_stream (orb, file, ev);

   fclose (file);

   return obj;
}

/**
 *
 */
static
void
client_run (Echo              echo_service,
            CORBA_Environment *ev)
{
   char filebuffer[1024+1];

   g_print ("Type message to the server\n"
            "a single dot in line will terminate input\n");

   while (fgets (filebuffer, 1024, stdin))
   {
      if (filebuffer[0] == '.' && filebuffer[1] == '\n')
         break;

      /* weg mit den Zeilenumbruch */
      filebuffer[strlen (filebuffer)-1] = '\0';

      /* endlich...wir nutzen die echoString Methode welche in echo.h
       * definiert ist welche wiederum aus der echo.idl erzeugt wurde. */

      Echo_echoString (echo_service, filebuffer, ev);
      if (raised_exception (ev))
         return;
   }
}

/*
 *
 */
int
main (int argc, char *argv[])
{
   CORBA_char        filename[] = "echo.ref";
   Echo              echo_service = CORBA_OBJECT_NIL;
   CORBA_Environment ev[1];

   CORBA_exception_init (ev);

   client_init (&argc, argv, &global_orb, ev);
   abort_if_exception (ev, "init failed");

   g_print ("Reading service reference from file \"%s\"\n", filename);

   echo_service = (Echo) client_import_service_from_file (global_orb,
                                                          filename,
                                                          ev);
   abort_if_exception (ev, "import service failed");

   client_run (echo_service, ev);
   abort_if_exception (ev, "service not reachable");

   client_cleanup (global_orb, echo_service, ev);
   abort_if_exception (ev, "cleanup failed");

   exit(0);
}
