/*
 *
 * echo-server Programm. Hacked aus der Echo test Umgebung von
 * <birney@sanger.ac.uk>, ORBit2 update durch Frank Rehberger
 * <F.Rehberger@xtradyne.de>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <orbit/orbit.h>

#include "echo-skelimpl.c"

static CORBA_ORB           global_orb = CORBA_OBJECT_NIL;   /* globaler orb */
static PortableServer_POA  root_poa   = CORBA_OBJECT_NIL;   /* root POA */

/**
 * auf Ausnahme (exception) testen
 */
static
gboolean
raised_exception(CORBA_Environment *ev)
{
   return ((ev)->_major != CORBA_NO_EXCEPTION);
}

/**
 * falls eine Ausnahme aufgetreten ist beendet diese Funktion den Prozess
 */
static
void
abort_if_exception(CORBA_Environment *ev, const char* mesg)
{
   if (raised_exception (ev)) {
      g_error ("%s %s", mesg, CORBA_exception_id (ev));
      CORBA_exception_free (ev);
      abort();
   }
}

/* Signalehandle um den Prozess sauber zu beenden falls KILL oder TERM
 * auftritt */
static
void
server_shutdown (int sig)
{
   CORBA_Environment    local_ev[1];

   CORBA_exception_init (local_ev);

   if (global_orb != CORBA_OBJECT_NIL)
   {
      CORBA_ORB_shutdown (global_orb, FALSE, local_ev);
      abort_if_exception (local_ev, "caught exception");
   }
}

/*
 * Initialisiert den ORB @orb durch die @argv Parameter. Für jeden ORBit
 * Parameter der aus @argv benutzt wird, wird der @argc_ptr um eins
 * verringert. Der Signal-Handler wird eingerichtet, so daß er
 * echo_client_shutdown aufruft, wenn die Signale SIGINT oder SIGTERM auftreten.
 * Wenn ein Fehler auftritt zeigt @ev nach der Rückkehr auf ein Exception
 * Objekt.
 */
#define  ORB_INIT_OK                   0x00
#define  ORB_INIT_FAILED_ORB           0x01
#define  ORB_INIT_FAILED_POA           0x02
#define  ORB_INIT_FAILED_POA_MANAGER   0x03
#define  ORB_INIT_FAILED_ACTIVATION    0x04

static
void
server_init (int                 *argc_ptr,
             char                *argv[],
             CORBA_ORB           *orb,
             PortableServer_POA  *poa,
             CORBA_Environment   *ev)
{
   PortableServer_POAManager  poa_manager = CORBA_OBJECT_NIL;
   CORBA_Environment          local_ev[1];
   int                        orb_init_fail_code = ORB_INIT_OK;

   CORBA_exception_init (local_ev);

   /* Signalhandler einrichten */
   signal (SIGINT,   server_shutdown);
   signal (SIGTERM,  server_shutdown);

   /* Object Request Broker (ORB) erzeugen */
   (*orb) = CORBA_ORB_init (argc_ptr, argv, "orbit_local_mt_orb", ev);
   if (raised_exception (ev))
      orb_init_fail_code = ORB_INIT_FAILED_ORB;

   if (orb_init_fail_code == ORB_INIT_OK)
   {
      (*poa) = (PortableServer_POA)
               CORBA_ORB_resolve_initial_references (*orb, "RootPOA", ev);
      if (raised_exception (ev))
         orb_init_fail_code = ORB_INIT_FAILED_POA;
   }

   if (orb_init_fail_code == ORB_INIT_OK)
   {
      poa_manager = (PortableServer_POAManager)
                    PortableServer_POA__get_the_POAManager (*poa, ev);
      if (raised_exception (ev))
         orb_init_fail_code = ORB_INIT_FAILED_POA_MANAGER;
   }

   if (orb_init_fail_code == ORB_INIT_OK)
   {
      PortableServer_POAManager_activate (poa_manager, ev);
      if (raised_exception (ev))
         orb_init_fail_code = ORB_INIT_FAILED_ACTIVATION;
   }

   switch (orb_init_fail_code)
   {
      case ORB_INIT_FAILED_ACTIVATION:
      case ORB_INIT_FAILED_POA_MANAGER:
         CORBA_Object_release ((CORBA_Object) poa_manager, ev);
      case ORB_INIT_FAILED_POA:
         CORBA_ORB_destroy (*orb, local_ev);
      case ORB_INIT_FAILED_ORB:
         break;
   }

   return;
}

/**
 * Betrete Hauptschleife. @orb bearbeitet ankommende Anfragen und verteilt Sie
 * auf Dienste.
 * Wenn ein Fehler auftritt zeigt @ev nach der Rückkehr auf ein Exception
 * Objekt.
 */
static
void
server_run (CORBA_ORB            orb,
            CORBA_Environment    *ev)
{
   /* eintritt in Hauptschleife bis ein SIGINT oder SIGTERM auftritt */

   CORBA_ORB_run (orb, ev);
   if (raised_exception (ev))
      return;

   /* falls ein Benutzer ein SIGINT oder SIGTERM erzeugt wird 
    * CORBA_ORB_shutdown aufgerufen und CORBA_ORB_run ordnungsgemäß beendet.*/
}

/**
 * gibt das Dienstobjekt @servant frei und zerstört anschließend @orb.
 * Wenn ein Fehler auftritt zeigt @ev nach der Rückkehr auf ein Exception
 * Objekt.
 */
static
void
server_cleanup (CORBA_ORB           orb,
                PortableServer_POA  poa,
                CORBA_Object        servant,
                CORBA_Environment   *ev)
{
   PortableServer_ObjectId    *objid = NULL;

   objid = (PortableServer_ObjectId*)
           PortableServer_POA_reference_to_id (poa, servant, ev);
   if (raised_exception (ev))
      return;

   /* Servant: Deaktivierung - ruft __destroy Destruktor auf */
   PortableServer_POA_deactivate_object (poa, objid, ev);
   if (raised_exception (ev))
      return;

   PortableServer_POA_destroy (poa, TRUE, FALSE, ev);
   if (raised_exception (ev))
      return;

   CORBA_free (objid);

   CORBA_Object_release ((CORBA_Object) poa, ev);
   if (raised_exception (ev))
      return;

   CORBA_Object_release (servant, ev);
   if (raised_exception (ev))
      return;

   /* ORB herunterfahren */
   if (orb != CORBA_OBJECT_NIL)
   {
      /* versuche den orb zu zerstören.. */
      CORBA_ORB_destroy (orb, ev);
      if (raised_exception (ev))
         return;
   }
}

/**
 * Erzeugt den Dienst und registriert ihn im ORB @orb. Der ORB verteilt 
 * ankommende Anfragen an die spezifischen Dienst-Objekte. @return ist 
 * die Objekt-Referenz.
 * Wenn ein Fehler auftritt zeigt @ev nach der Rückkehr auf ein Exception
 * Objekt.
 */
static
CORBA_Object
server_activate_service (CORBA_ORB           orb,
                         PortableServer_POA  poa,
                         CORBA_Environment   *ev)
{
   Echo  ref = CORBA_OBJECT_NIL;

   ref = impl_Echo__create (poa, ev);
   if (raised_exception (ev))
      return CORBA_OBJECT_NIL;

   return ref;
}

/**
 *
 */
static
CORBA_Object
server_export_service_to_stream (CORBA_ORB          orb,
                                 CORBA_Object       servant,
                                 FILE               *stream,
                                 CORBA_Environment  *ev)
{
   CORBA_char *objref = NULL;

   objref = CORBA_ORB_object_to_string (orb, servant, ev);
   if (raised_exception (ev))
      return;

   /* ior auf Stream ausgeben */
   fprintf (stream, "%s", objref);
   fflush (stream);

   CORBA_free (objref);
}

/**
 *
 */
static
CORBA_Object
server_export_service_to_file (CORBA_ORB            orb,
                               CORBA_Object         servant,
                               char                 *filename,
                               CORBA_Environment    *ev)
{
   FILE           *file   = NULL;

   file = fopen (filename, "w");
   if (file == NULL)
      g_error ("could not open %s\n", filename);

   /* schreibe ior in Datei */
   server_export_service_to_stream (orb, servant, file, ev);

   fclose (file);
}

/**
 * 
 */
int
main (int argc, char *argv[])
{
   CORBA_Object      servant = CORBA_OBJECT_NIL;
   CORBA_char        filename[] = "echo.ref";
   CORBA_Environment ev[1];

   CORBA_exception_init (ev);

   server_init (&argc, argv, &global_orb, &root_poa, ev);
   abort_if_exception (ev, "failed ORB init");

   servant = server_activate_service (global_orb, root_poa, ev);
   abort_if_exception (ev, "failed activating service");

   g_print ("Writing service reference to: %s\n\n", filename);

   server_export_service_to_file (global_orb,
                                  servant,
                                  filename,
                                  ev);
   abort_if_exception (ev, "failed exporting IOR");

   server_run (global_orb, ev);
   abort_if_exception (ev, "failed entering main loop");

   server_cleanup (global_orb, root_poa, servant, ev);
   abort_if_exception (ev, "failed cleanup");

   exit(0);
}

